//////////////////////////////////////////////////////////////////////////////////////////////////
// i2c.c - funkcje podstawowe programowego interfejsu i2c, bez moliwoci przejcia magistrali
//			(tylko jeden master)
// 
// Autor: Radosaw Koppel          Kompilator: WinAVR 20050214
//////////////////////////////////////////////////////////////////////////////////////////////////
// Komenta dodatkowy:
//   Linie SCL naley skonfigurowa jako wyjciow, stan pocztkowy = 1
//   Linie SDA naley skonfigurowa jako wejciow z podciganiem do +
//////////////////////////////////////////////////////////////////////////////////////////////////
#include <avr\io.h>
#include <inttypes.h>

#include "harddef.h"
#include "makra.h"
#include "i2c.h"

// Wyliczenie czasu opnienia powkowego i wiartkowego (cykle)
#define I2C_nhalf (F_CPU/I2C_SPEED/2)

// Funkcja duszych opnie
#if I2C_nhalf < 3
	// Nic
#elif I2C_nhalf < 8
	static void i2c_xdelay(void)
	{
		NOP();
	}
#else
	#define I2C_delayloops (1+(I2C_nhalf-8)/3)
	#if I2C_delayloops > 255
		#error Przyspiesz - bo sie nie wyrabiam ;)
	#endif
	static void i2c_xdelay(void)
	{
		asm volatile( \
			"delayus8_loop%=: \n\t"\
			"dec %[ticks] \n\t"\
			"brne delayus8_loop%= \n\t"\
		: :[ticks]"r"(I2C_delayloops) );
	}
#endif //I2C_nhalf >= 3

// Opnienia dla I2C
static inline void i2c_hdelay(void)
{
#if I2C_nhalf < 1
	return; // To jest funkcja inline, jeli skada si tylko z "return" jest usuwana podczas optymalizacji
#elif I2C_nhalf < 2
	NOP();
#elif I2C_nhalf < 3
	asm volatile(
		"rjmp exit%=\n\t"
		"exit%=:\n\t"::);
#else
	i2c_xdelay();
#endif	
}

// Ustawienie i zerowanie wyjcia
static inline void i2c_sdaset(void)
{
	DDR(I2C_SDAPORT) &= ~(1<<I2C_SDA);
	PORT(I2C_SDAPORT) |= 1<<I2C_SDA;
}
static inline void i2c_sdaclear(void)
{
	PORT(I2C_SDAPORT) &= ~(1<<I2C_SDA);
	DDR(I2C_SDAPORT) |= 1<<I2C_SDA;
}
// Pobieranie danej z wyprowadzenia portu
// Zwraca: bajt bdcy odpowiednikiem fizycznego stanu portu, z wyzerowanaymi wszystkimi bitami poza sda
static inline uint8_t i2c_sdaget(void)
{
	return PIN(I2C_SDAPORT) & (1<<I2C_SDA);
}

// Zerowanie i ustawianie zegara
static inline void i2c_sclset(void)
{
	PORT(I2C_SCLPORT) |= 1<<I2C_SCL;
}
static inline void i2c_sclclear(void)
{
	PORT(I2C_SCLPORT) &= ~(1<<I2C_SCL);
}

// Warunek startu
void i2c_start(void)
{
	// Konieczne jeli chc uy start bez stop
	i2c_sdaset();
	i2c_hdelay();
	i2c_sclset();
	i2c_hdelay();
	// Normalna sekwencja startu
	i2c_sdaclear();
	i2c_hdelay();
	i2c_sclclear();
}

// Warunek stop
void i2c_stop(void)
{
	i2c_sdaclear();
	i2c_hdelay();
	i2c_sclset();
	i2c_hdelay();
	i2c_sdaset();
	i2c_hdelay();
}

// Transmisja bajtu
// Zwraca: 0 jeli byo ack, warto 1<<I2C_SDA jeli nie byo ACK
uint8_t i2c_send(uint8_t data)
{
	uint8_t n;
	
	for(n=8; n>0; --n)
	{
		if(data & 0x80)
			i2c_sdaset();
		else
			i2c_sdaclear();
		data <<= 1;
		i2c_hdelay();
		i2c_sclset();
		i2c_hdelay();
		i2c_sclclear();
	}
	//  ack
	i2c_sdaset();
	i2c_hdelay();
	i2c_sclset();
	i2c_hdelay();
	n = i2c_sdaget();
	i2c_sclclear();
	
	return n;
}

// Pobranie bajtu
uint8_t i2c_get(uint8_t ack)
{
	uint8_t n, temp=0;
	
	i2c_sdaset();
	for(n=8; n>0; --n)
	{
		i2c_hdelay();
		i2c_sclset();
		i2c_hdelay();
		temp<<=1;
		if(i2c_sdaget())
			temp++;
		i2c_sclclear();
	}
	// ack
	if(ack == I2C_ACK)
		i2c_sdaclear();
	else
		i2c_sdaset();
	i2c_hdelay();
	i2c_sclset();
	i2c_hdelay();
	i2c_sclclear();
	
	return temp;
}
